import pytest

from aiodocker.exceptions import DockerError, DockerContainerError


async def _validate_hello(container):
    try:
        await container.start()
        response = await container.wait()
        assert response['StatusCode'] == 0
        logs = await container.log(stdout=True)
        assert logs == ['hello']

        with pytest.raises(TypeError):
            await container.log()
    finally:
        await container.delete(force=True)


@pytest.mark.asyncio
async def test_run_existing_container(docker):
    name = "alpine:latest"
    await docker.pull(name)
    container = await docker.containers.run(
        config={
            'Cmd': ['-c', 'echo hello'],
            'Entrypoint': 'sh',
            'Image': name
        }
    )

    await _validate_hello(container)


@pytest.mark.asyncio
async def test_run_container_with_missing_image(docker):
    name = "alpine:latest"
    try:
        await docker.images.delete(name)
    except DockerError as e:
        if e.status == 404:
            pass  # already missing, pass
        else:
            raise

    # should automatically pull the image
    container = await docker.containers.run(
        config={
            'Cmd': ['-c', 'echo hello'],
            'Entrypoint': 'sh',
            'Image': name
        }
    )

    await _validate_hello(container)


@pytest.mark.asyncio
async def test_run_failing_start_container(docker):
    name = "alpine:latest"
    try:
        await docker.images.delete(name)
    except DockerError as e:
        if e.status == 404:
            pass  # already missing, pass
        else:
            raise

    with pytest.raises(DockerContainerError) as e_info:
        await docker.containers.run(
            config={
                # we want to raise an error
                # `executable file not found`
                'Cmd': ['pyton', 'echo hello'],
                'Image': name
            }
        )

    assert e_info.value.container_id
    # This container is created but not started!
    # We should delete it afterwards.
    cid = e_info.value.container_id
    container = docker.containers.container(cid)
    await container.delete()



import asyncio
import json
import logging
import os
from pathlib import Path
import re
import ssl

import aiohttp
from yarl import URL

from .utils import httpize, parse_result

# Sub-API classes
from .containers import DockerContainers, DockerContainer
from .events import DockerEvents
from .exceptions import DockerError
from .images import DockerImages
from .logs import DockerLog
from .swarm import DockerSwarm
from .services import DockerServices
from .tasks import DockerTasks
from .volumes import DockerVolumes, DockerVolume
from .nodes import DockerSwarmNodes

__all__ = (
    'Docker',
    'DockerContainers', 'DockerContainer',
    'DockerEvents',
    'DockerError',
    'DockerImages',
    'DockerLog',
    'DockerSwarm',
    'DockerServices',
    'DockerTasks',
    'DockerVolumes', 'DockerVolume',
    'DockerSwarmNodes',
)

log = logging.getLogger(__name__)

_sock_search_paths = [
    Path('/run/docker.sock'),
    Path('/var/run/docker.sock'),
]

_rx_version = re.compile(r'^v\d+\.\d+$')
_rx_tcp_schemes = re.compile(r'^(tcp|http)://')


class Docker:
    def __init__(self,
                 url=None,
                 connector=None,
                 session=None,
                 ssl_context=None,
                 api_version='v1.30'):

        docker_host = url  # rename
        if docker_host is None:
            docker_host = os.environ.get('DOCKER_HOST', None)
        if docker_host is None:
            for sockpath in _sock_search_paths:
                if sockpath.is_socket():
                    docker_host = 'unix://' + str(sockpath)
                    break
        self.docker_host = docker_host

        assert _rx_version.search(api_version) is not None, \
            'Invalid API version format'
        self.api_version = api_version

        if docker_host is None:
            raise ValueError(
                "Missing valid docker_host."
                "Either DOCKER_HOST or local sockets are not available."
                )

        if connector is None:
            if _rx_tcp_schemes.search(docker_host):
                if os.environ.get('DOCKER_TLS_VERIFY', '0') == '1':
                    ssl_context = self._docker_machine_ssl_context()
                    docker_host = _rx_tcp_schemes.sub('https://', docker_host)
                else:
                    ssl_context = None
                connector = aiohttp.TCPConnector(ssl_context=ssl_context)
                self.docker_host = docker_host
            elif docker_host.startswith('unix://'):
                connector = aiohttp.UnixConnector(docker_host[7:])
                # dummy hostname for URL composition
                self.docker_host = "unix://localhost"
            else:
                raise ValueError('Missing protocol scheme in docker_host.')
        self.connector = connector
        if session is None:
            session = aiohttp.ClientSession(connector=self.connector)
        self.session = session

        self.events = DockerEvents(self)
        self.containers = DockerContainers(self)
        self.swarm = DockerSwarm(self)
        self.services = DockerServices(self)
        self.tasks = DockerTasks(self)
        self.images = DockerImages(self)
        self.volumes = DockerVolumes(self)
        self.nodes = DockerSwarmNodes(self)

        # legacy aliases
        self.pull = self.images.pull
        self.push = self.images.push

    async def close(self):
        await self.events.stop()
        await self.session.close()

    async def auth(self, **credentials):
        response = await self._query_json(
            "auth", "POST",
            data=credentials,
        )
        return response

    async def version(self):
        data = await self._query_json("version")
        return data

    def _canonicalize_url(self, path):
        return URL("{self.docker_host}/{self.api_version}/{path}"
                   .format(self=self, path=path))

    async def _query(self, path, method='GET', *,
                     params=None, data=None, headers=None,
                     timeout=None):
        '''
        Get the response object by performing the HTTP request.
        The caller is responsible to finalize the response object.
        '''
        url = self._canonicalize_url(path)
        if headers and 'content-type' not in headers:
            headers['content-type'] = 'application/json'
        try:
            response = await self.session.request(
                method, url,
                params=httpize(params),
                headers=headers,
                data=data,
                timeout=timeout)
        except asyncio.TimeoutError:
            raise
        if (response.status // 100) in [4, 5]:
            what = await response.read()
            content_type = response.headers.get('content-type', '')
            response.close()
            if content_type == 'application/json':
                raise DockerError(response.status,
                                  json.loads(what.decode('utf8')))
            else:
                raise DockerError(response.status,
                                  {"message": what.decode('utf8')})
        return response

    async def _query_json(self, path, method='GET', *,
                          params=None, data=None, headers=None,
                          timeout=None):
        '''
        A shorthand of _query() that treats the input as JSON.
        '''
        if headers is None:
            headers = {}
        headers['content-type'] = 'application/json'
        if not isinstance(data, (str, bytes)):
            data = json.dumps(data)
        response = await self._query(
            path, method,
            params=params, data=data, headers=headers,
            timeout=timeout)
        data = await parse_result(response)
        return data

    async def _websocket(self, path, **params):
        if not params:
            params = {
                'stdin': True,
                'stdout': True,
                'stderr': True,
                'stream': True
            }
        url = self._canonicalize_url(path)
        # ws_connect() does not have params arg.
        url = url.with_query(httpize(params))
        ws = await self.session.ws_connect(
            url,
            protocols=['chat'],
            origin='http://localhost',
            autoping=True,
            autoclose=True)
        return ws

    @staticmethod
    def _docker_machine_ssl_context():
        '''
        Create a SSLContext object using DOCKER_* env vars.
        '''
        context = ssl.SSLContext(ssl.PROTOCOL_TLS)
        context.set_ciphers(ssl._RESTRICTED_SERVER_CIPHERS)
        certs_path = os.environ.get('DOCKER_CERT_PATH', None)
        if certs_path is None:
            raise ValueError("Cannot create ssl context, "
                             "DOCKER_CERT_PATH is not set!")
        certs_path = Path(certs_path)
        context.load_verify_locations(cafile=certs_path / 'ca.pem')
        context.load_cert_chain(certfile=certs_path / 'cert.pem',
                                keyfile=certs_path / 'key.pem')
        return context



import asyncio
import base64
import codecs
from io import BytesIO
import sys
from typing import (
    Any, Iterable, Optional, Union,
    MutableMapping, Mapping, Tuple,
    BinaryIO, IO,
)
import tempfile
import tarfile
import json


async def parse_result(response, response_type=None, *,
                       encoding='utf-8'):
    '''
    Convert the response to native objects by the given response type
    or the auto-detected HTTP content-type.
    It also ensures release of the response object.
    '''
    if response_type is None:
        ct = response.headers.get('content-type')
        if ct is None:
            raise TypeError('Cannot auto-detect respone type '
                            'due to missing Content-Type header.')
        main_type, sub_type, extras = parse_content_type(ct)
        if sub_type == 'json':
            response_type = 'json'
        elif sub_type == 'x-tar':
            response_type = 'tar'
        elif (main_type, sub_type) == ('text', 'plain'):
            response_type = 'text'
            encoding = extras.get('charset', encoding)
        else:
            raise TypeError("Unrecognized response type: {ct}"
                            .format(ct=ct))
    if 'tar' == response_type:
        what = await response.read()
        return tarfile.open(mode='r', fileobj=BytesIO(what))
    if 'json' == response_type:
        data = await response.json(encoding=encoding)
    elif 'text' == response_type:
        data = await response.text(encoding=encoding)
    else:
        data = await response.read()
    return data


def parse_content_type(ct: str) -> Tuple[str, str, Mapping[str, str]]:
    '''
    Decompose the value of HTTP "Content-Type" header into
    the main/sub MIME types and other extra options as a dictionary.
    All parsed values are lower-cased automatically.
    '''
    pieces = ct.split(';')
    try:
        main_type, sub_type = pieces[0].split('/')
    except ValueError:
        msg = 'Invalid mime-type component: "{0}"'.format(pieces[0])
        raise ValueError(msg)
    if len(pieces) > 1:
        options = {}
        for opt in pieces[1:]:
            opt = opt.strip()
            if not opt:
                continue
            try:
                k, v = opt.split('=', 1)
            except ValueError:
                msg = 'Invalid option component: "{0}"'.format(opt)
                raise ValueError(msg)
            else:
                options[k.lower()] = v.lower()
    else:
        options = {}
    return main_type.lower(), sub_type.lower(), options


def identical(d1, d2):
    if type(d1) != type(d2):
        return False

    if isinstance(d1, dict):
        keys = set(d1.keys()) | set(d2.keys())
        for key in keys:
            if not identical(d1.get(key, {}), d2.get(key, {})):
                return False
        return True

    if isinstance(d1, list):
        if len(d1) != len(d2):
            return False

        pairs = zip(d1, d2)
        return all((identical(x, y) for (x, y) in pairs))

    return d1 == d2


_true_strs = frozenset(['true', 'yes', 'y', '1'])
_false_strs = frozenset(['false', 'no', 'n', '0'])


def human_bool(s) -> bool:
    if isinstance(s, str):
        if s.lower() in _true_strs:
            return True
        if s.lower() in _false_strs:
            return False
        raise ValueError('Cannot interpret {s!r} as boolean.'.format(s=s))
    else:
        return bool(s)


def httpize(d: Optional[Mapping]) -> Mapping[str, Any]:
    if d is None:
        return None
    converted = {}
    for k, v in d.items():
        if isinstance(v, bool):
            v = '1' if v else '0'
        if not isinstance(v, str):
            v = str(v)
        converted[k] = v
    return converted


class _DecodeHelper:
    """
    Decode logs from the Docker Engine
    """

    def __init__(self, generator, encoding):
        self._gen = generator.__aiter__()
        self._decoder = codecs.getincrementaldecoder(encoding)(errors='ignore')
        self._flag = False

    def __aiter__(self):
        return self

    # to make it compatible with Python 3.5.0 and 3.5.2
    # https://www.python.org/dev/peps/pep-0492/#api-design-and-implementation-revisions
    if sys.version_info <= (3, 5, 2):
        __aiter__ = asyncio.coroutine(__aiter__)

    async def __anext__(self):
        if self._flag:
            raise StopAsyncIteration

        # we catch StopAsyncIteration from self._gen
        # because we need to close the decoder
        # then we raise StopAsyncIteration checking self._flag
        try:
            stream = await self._gen.__anext__()
        except StopAsyncIteration:
            self._flag = True
            stream_decoded = self._decoder.decode(b'', final=True)
            if stream_decoded:
                return stream_decoded
            raise StopAsyncIteration
        else:
            return self._decoder.decode(stream)


def clean_map(obj: Mapping[Any, Any]) -> Mapping[Any, Any]:
    """
    Return a new copied dictionary without the keys with ``None`` values from
    the given Mapping object.
    """
    return {k: v for k, v in obj.items() if v is not None}


def format_env(key, value: Union[None, bytes, str]) -> str:
    """
    Formats envs from {key:value} to ['key=value']
    """
    if value is None:
        return key
    if isinstance(value, bytes):
        value = value.decode('utf-8')

    return "{key}={value}".format(key=key, value=value)


def clean_networks(networks: Optional[Iterable[str]]=None) -> Iterable[str]:
    """
    Cleans the values inside `networks`
    Returns a new list
    """
    if not networks:
        return networks
    if not isinstance(networks, list):
        raise TypeError('networks parameter must be a list.')

    result = []
    for n in networks:
        if isinstance(n, str):
            n = {'Target': n}
        result.append(n)
    return result


def clean_filters(filters: Optional[dict]=None) -> str:
    """
    Checks the values inside `filters`
    https://docs.docker.com/engine/api/v1.29/#operation/ServiceList
    Returns a new dictionary in the format `map[string][]string` jsonized
    """

    if filters and isinstance(filters, dict):
        for k, v in filters.items():
            if not isinstance(v, list):
                v = [v, ]
            filters[k] = v

    return json.dumps(filters)


def mktar_from_dockerfile(fileobject: BinaryIO) -> IO:
    """
    Create a zipped tar archive from a Dockerfile
    **Remember to close the file object**
    Args:
        fileobj: a Dockerfile
    Returns:
        a NamedTemporaryFile() object
    """

    f = tempfile.NamedTemporaryFile()
    t = tarfile.open(mode='w:gz', fileobj=f)

    if isinstance(fileobject, BytesIO):
        dfinfo = tarfile.TarInfo('Dockerfile')
        dfinfo.size = len(fileobject.getvalue())
        fileobject.seek(0)
    else:
        dfinfo = t.gettarinfo(fileobj=fileobject, arcname='Dockerfile')

    t.addfile(dfinfo, fileobject)
    t.close()
    f.seek(0)
    return f


def compose_auth_header(auth: Union[MutableMapping, str, bytes],
                        registry_addr: Optional[str]=None) -> str:
    """
    Validate and compose base64-encoded authentication header
    with an optional support for parsing legacy-style "user:password"
    strings.

    Args:
        auth: Authentication information
        registry_addr: An address of the registry server

    Returns:
        A base64-encoded X-Registry-Auth header value
    """
    if isinstance(auth, Mapping):
        # Validate the JSON format only.
        if 'identitytoken' in auth:
            pass
        elif 'auth' in auth:
            return compose_auth_header(auth['auth'], registry_addr)
        else:
            if registry_addr:
                auth['serveraddress'] = registry_addr
        auth_json = json.dumps(auth).encode('utf-8')
        auth = base64.b64encode(auth_json).decode('ascii')
    elif isinstance(auth, (str, bytes)):
        # Parse simple "username:password"-formatted strings
        # and attach the server address specified.
        if isinstance(auth, bytes):
            auth = auth.decode('utf-8')
        s = base64.b64decode(auth)
        username, passwd = s.split(b':', 1)
        config = {
            "username": username.decode('utf-8'),
            "password": passwd.decode('utf-8'),
            "email": None,
            "serveraddress": registry_addr,
        }
        auth_json = json.dumps(config).encode('utf-8')
        auth = base64.b64encode(auth_json).decode('ascii')
    else:
        raise TypeError(
            "auth must be base64 encoded string/bytes or a dictionary")
    return auth



import json
from typing import Optional, Dict


class DockerSwarm(object):
    def __init__(self, docker):
        self.docker = docker

    async def init(self,
                   *,
                   advertise_addr: Optional[str]=None,
                   listen_addr: str="0.0.0.0:2377",
                   force_new_cluster: bool=False,
                   swarm_spec: Optional[Dict]=None
                   ) -> str:
        """
        Initialize a new swarm

        Args:
            ListenAddr: listen address used for inter-manager communication
            AdvertiseAddr: address advertised to other nodes.
            ForceNewCluster: Force creation of a new swarm.
            SwarmSpec: User modifiable swarm configuration.

        Returns:
            id of the swarm node
        """

        data = {
            'AdvertiseAddr': advertise_addr,
            'ListenAddr': listen_addr,
            'ForceNewCluster': force_new_cluster,
            'Spec': swarm_spec,
        }

        data_str = json.dumps(data)

        response = await self.docker._query(
            "swarm/init",
            method='POST',
            headers={"content-type": "application/json", },
            data=data_str
        )

        return response

    async def inspect(self) -> Dict:
        """
        Inspect a swarm

        Returns:
            Info about the swarm
        """

        response = await self.docker._query_json(
            "swarm",
            method='GET',
        )

        return response

    async def leave(self, *, force: bool=False) -> bool:
        """
        Leave a swarm

        Args:
            force: force to leave the swarm even if the node is a master

        """

        params = {"force": force}

        await self.docker._query(
            "swarm/leave",
            method='POST',
            params=params
        )

        return True



import asyncio
import uuid
from distutils.version import StrictVersion
from os import environ as ENV
import traceback

import pytest

from aiodocker.docker import Docker
from aiodocker.exceptions import DockerError


_api_versions = {
    "17.06": "v1.30",
}


def _random_name():
    return "aiodocker-" + uuid.uuid4().hex[:7]


@pytest.fixture(scope='session')
def random_name():
    yield _random_name

    # If some test cases have used randomly-named temporary images,
    # we need to clean up them!
    if ENV.get('CI', '') == 'true':
        # But inside the CI server, we don't need clean up!
        return
    event_loop = asyncio.get_event_loop()

    async def _clean():
        docker = Docker()
        images = await docker.images.list()
        for img in images:
            if img['RepoTags'] is None:
                continue
            try:
                if img['RepoTags'][0].startswith('aiodocker-'):
                    print('Deleting image id: {0}'.format(img['Id']))
                    await docker.images.delete(img['Id'], force=True)
            except DockerError as e:
                traceback.print_exc()
        await docker.close()

    event_loop.run_until_complete(_clean())


@pytest.fixture(scope='session')
def testing_images():
    # Prepare a small Linux image shared by most test cases.
    event_loop = asyncio.get_event_loop()

    async def _pull():
        docker = Docker()
        required_images = [
            'alpine:latest',
            'redis:latest',
            'redis:3.0.2',
            'redis:4.0',
            'python:3.6.1-alpine',
        ]
        for img in required_images:
            try:
                await docker.images.get(img)
            except DockerError as e:
                assert e.status == 404
                print('Pulling "{img}" for the testing session...'
                      .format(img=img))
                await docker.pull(img)
        await docker.close()

    event_loop.run_until_complete(_pull())


@pytest.fixture
def docker(event_loop, testing_images):
    kwargs = {}
    if "DOCKER_VERSION" in ENV:
        kwargs["api_version"] = _api_versions[ENV["DOCKER_VERSION"]]
    docker = Docker(**kwargs)
    yield docker

    async def _finalize():
        await docker.close()
    event_loop.run_until_complete(_finalize())


@pytest.fixture
def requires_api_version(docker):

    def check(version, reason):
        if StrictVersion(docker.api_version[1:]) < StrictVersion(version[1:]):
            pytest.skip(reason)

    yield check


@pytest.fixture
def swarm(event_loop, docker):
    assert event_loop.run_until_complete(docker.swarm.init())
    yield docker
    assert event_loop.run_until_complete(docker.swarm.leave(force=True))


@pytest.fixture
def shell_container(event_loop, docker):
    container = None
    config = {
        "Cmd": ["/bin/ash"],
        "Image": "alpine:latest",
        "AttachStdin": False,
        "AttachStdout": False,
        "AttachStderr": False,
        "Tty": True,
        "OpenStdin": True,
    }

    async def _spawn():
        nonlocal container
        container = await docker.containers.create_or_replace(
            config=config,
            name='aiodocker-testing-shell')
        await container.start()
    event_loop.run_until_complete(_spawn())

    yield container

    async def _delete():
        nonlocal container
        await container.delete(force=True)
    event_loop.run_until_complete(_delete())


@pytest.fixture
def redis_container(event_loop, docker):
    container = None
    config = {
        "Image": "redis:latest",
        "PublishAllPorts": True,
    }

    async def _spawn():
        nonlocal container
        container = await docker.containers.create_or_replace(
            config=config,
            name='aiodocker-testing-redis')
        await container.start()
    event_loop.run_until_complete(_spawn())

    yield container

    async def _delete():
        nonlocal container
        await container.delete(force=True)
    event_loop.run_until_complete(_delete())



import json
from typing import (
    Optional, Union,
    List, MutableMapping, Mapping,
    BinaryIO,
)
from .utils import clean_map, compose_auth_header
from .jsonstream import json_stream_result


class DockerImages(object):
    def __init__(self, docker):
        self.docker = docker

    async def list(self, **params) -> Mapping:
        """
        List of images
        """
        response = await self.docker._query_json(
            "images/json", "GET",
            params=params,
        )
        return response

    async def get(self, name: str) -> Mapping:
        """
        Return low-level information about an image

        Args:
            name: name of the image
        """
        response = await self.docker._query_json(
            "images/{name}/json".format(name=name),
        )
        return response

    async def history(self, name: str) -> Mapping:
        response = await self.docker._query_json(
            "images/{name}/history".format(name=name),
        )
        return response

    async def pull(self, from_image: str, *,
                   auth: Optional[Union[MutableMapping, str, bytes]]=None,
                   tag: Optional[str]=None,
                   repo: Optional[str]=None,
                   stream: bool=False) -> Mapping:
        """
        Similar to `docker pull`, pull an image locally

        Args:
            fromImage: name of the image to pull
            repo: repository name given to an image when it is imported
            tag: if empty when pulling an image all tags
                 for the given image to be pulled
            auth: special {'auth': base64} pull private repo
        """
        image = from_image  # TODO: clean up
        params = {
            'fromImage': image,
        }
        headers = {}
        if repo:
            params['repo'] = repo
        if tag:
            params['tag'] = tag
        if auth is not None:
            registry, has_registry_host, _ = image.partition('/')
            if not has_registry_host:
                raise ValueError('Image should have registry host '
                                 'when auth information is provided')
            # TODO: assert registry == repo?
            headers['X-Registry-Auth'] = compose_auth_header(auth, registry)
        response = await self.docker._query(
            "images/create",
            "POST",
            params=params,
            headers=headers,
        )
        return (await json_stream_result(response, stream=stream))

    async def push(self, name: str, *,
                   auth: Union[MutableMapping, str, bytes]=None,
                   tag: Optional[str]=None,
                   stream: bool=False) -> Mapping:
        params = {}
        headers = {
            # Anonymous push requires a dummy auth header.
            'X-Registry-Auth': 'placeholder',
        }
        if tag:
            params['tag'] = tag
        if auth is not None:
            registry, has_registry_host, _ = name.partition('/')
            if not has_registry_host:
                raise ValueError('Image should have registry host '
                                 'when auth information is provided')
            headers['X-Registry-Auth'] = compose_auth_header(auth, registry)
        response = await self.docker._query(
            "images/{name}/push".format(name=name),
            "POST",
            params=params,
            headers=headers,
        )
        return (await json_stream_result(response, stream=stream))

    async def tag(self, name: str, repo: str, *,
                  tag: Optional[str]=None) -> bool:
        """
        Tag the given image so that it becomes part of a repository.

        Args:
            repo: the repository to tag in
            tag: the name for the new tag
        """
        params = {"repo": repo}

        if tag:
            params["tag"] = tag

        await self.docker._query(
            "images/{name}/tag".format(name=name),
            "POST",
            params=params,
            headers={"content-type": "application/json"},
        )
        return True

    async def delete(self, name: str, *, force: bool=False,
                     noprune: bool=False) -> List:
        """
        Remove an image along with any untagged parent
        images that were referenced by that image

        Args:
            name: name/id of the image to delete
            force: remove the image even if it is being used
                   by stopped containers or has other tags
            noprune: don't delete untagged parent images

        Returns:
            List of deleted images
        """
        params = {'force': force, 'noprune': noprune}
        response = await self.docker._query_json(
            "images/{name}".format(name=name),
            "DELETE",
            params=params,
        )
        return response

    async def build(self, *,
                    remote: Optional[str]=None,
                    fileobj: Optional[BinaryIO]=None,
                    path_dockerfile: Optional[str]=None,
                    tag: Optional[str]=None,
                    quiet: bool=False,
                    nocache: bool=False,
                    buildargs: Optional[Mapping]=None,
                    pull: bool=False,
                    rm: bool=True,
                    forcerm: bool=False,
                    labels: Optional[Mapping]=None,
                    stream: bool=False,
                    encoding: Optional[str]=None) -> Mapping:
        """
        Build an image given a remote Dockerfile
        or a file object with a Dockerfile inside

        Args:
            path_dockerfile: path within the build context to the Dockerfile
            remote: a Git repository URI or HTTP/HTTPS context URI
            quiet: suppress verbose build output
            nocache: do not use the cache when building the image
            rm: remove intermediate containers after a successful build
            pull: downloads any updates to the FROM image in Dockerfiles
            encoding: set `Content-Encoding` for the file object your send
            forcerm: always remove intermediate containers, even upon failure
            labels: arbitrary key/value labels to set on the image
            fileobj: a tar archive compressed or not
        """

        local_context = None

        headers = {}

        params = {
            't': tag,
            'rm': rm,
            'q': quiet,
            'pull': pull,
            'remote': remote,
            'nocache': nocache,
            'forcerm': forcerm,
            'dockerfile': path_dockerfile,
        }

        if remote is None and fileobj is None:
            raise ValueError("You need to specify either remote or fileobj")

        if fileobj and remote:
            raise ValueError("You cannot specify both fileobj and remote")

        if fileobj and not encoding:
            raise ValueError("You need to specify an encoding")

        if remote is None and fileobj is None:
            raise ValueError("Either remote or fileobj needs to be provided.")

        if fileobj:
            local_context = fileobj.read()
            headers["content-type"] = "application/x-tar"

        if fileobj and encoding:
            headers['Content-Encoding'] = encoding

        if buildargs:
            params.update({'buildargs': json.dumps(buildargs)})

        if labels:
            params.update({'labels': json.dumps(labels)})

        response = await self.docker._query(
            "build",
            "POST",
            params=clean_map(params),
            headers=headers,
            data=local_context
        )

        return (await json_stream_result(response, stream=stream))



import json
from typing import Optional, Dict, List, Any, Union, AsyncIterator
from .multiplexed import multiplexed_result
from .utils import clean_map, clean_networks, clean_filters, format_env


class DockerServices(object):
    def __init__(self, docker):
        self.docker = docker

    async def list(self, *, filters: Optional[Dict]=None) -> List[Dict]:
        """
        Return a list of services

        Args:
            filters: a dict with a list of filters

        Available filters:
            id=<service id>
            label=<service label>
            mode=["replicated"|"global"]
            name=<service name>
        """

        params = {"filters": clean_filters(filters)}

        response = await self.docker._query_json(
            "services",
            method="GET",
            params=params
        )
        return response

    async def create(self,
                     task_template: Dict[str, Any],
                     *,
                     name: Optional[str]=None,
                     labels: Optional[List]=None,
                     mode: Optional[Dict]=None,
                     update_config: Optional[Dict]=None,
                     rollback_config: Optional[Dict]=None,
                     networks: Optional[List]=None,
                     endpoint_spec: Optional[Dict]=None
                     ) -> Dict[str, Any]:
        """
        Create a service

        Args:
            task_template: user modifiable task configuration
            name: name of the service
            labels: user-defined key/value metadata
            mode: scheduling mode for the service
            update_config: update strategy of the service
            rollback_config: rollback strategy of the service
            networks: array of network names/IDs to attach the service to
            endpoint_spec: ports to expose

        Returns:
            a dict with info of the created service
        """

        if "Image" not in task_template["ContainerSpec"]:
            raise KeyError(
                "Missing mandatory Image key in ContainerSpec"
            )

        # from {"key":"value"} to ["key=value"]
        if "Env" in task_template["ContainerSpec"]:
            task_template["ContainerSpec"]["Env"] = [
                format_env(k, v)
                for k, v in task_template["ContainerSpec"]["Env"].items()
            ]

        config = {
            "TaskTemplate": task_template,
            "Name": name,
            "Labels": labels,
            "Mode": mode,
            "UpdateConfig": update_config,
            "RollbackConfig": rollback_config,
            "Networks": clean_networks(networks),
            "EndpointSpec": endpoint_spec
        }

        data = json.dumps(clean_map(config))

        response = await self.docker._query_json(
            "services/create",
            method="POST",
            data=data,
        )
        return response

    async def update(self,
                     service_id: str,
                     version: str,
                     *,
                     image: str=None,
                     rollback: bool=False
                     ) -> bool:
        """
        Update a service.
        If rollback is True image will be ignored.

        Args:
            service_id: ID or name of the service.
            version: Version of the service that you want to update.
            rollback: Rollback the service to the previous service spec.

        Returns:
            True if successful.
        """
        if image is None and rollback is False:
            raise ValueError("You need to specify an image.")

        inspect_service = await self.inspect(service_id)
        spec = inspect_service["Spec"]

        if image is not None:
            spec['TaskTemplate']['ContainerSpec']['Image'] = image

        params = {
            "version": version,
        }
        if rollback is True:
            params["rollback"] = 'previous'

        data = json.dumps(clean_map(spec))

        await self.docker._query_json(
            "services/{service_id}/update".format(service_id=service_id),
            method="POST",
            data=data,
            params=params
        )
        return True

    async def delete(self, service_id: str) -> bool:
        """
        Remove a service

        Args:
            service_id: ID or name of the service

        Returns:
            True if successful
        """

        await self.docker._query(
            "services/{service_id}".format(service_id=service_id),
            method="DELETE",
        )
        return True

    async def inspect(self, service_id: str) -> Dict[str, Any]:
        """
        Inspect a service

        Args:
            service_id: ID or name of the service

        Returns:
            a dict with info about a service
        """

        response = await self.docker._query_json(
            "services/{service_id}".format(service_id=service_id),
            method="GET",
        )
        return response

    async def logs(self, service_id: str, *,
                   details: bool=False,
                   follow: bool=False,
                   stdout: bool=False,
                   stderr: bool=False,
                   since: int=0,
                   timestamps: bool=False,
                   is_tty: bool=False,
                   tail: str="all"
                   ) -> Union[str, AsyncIterator[str]]:
        """
        Retrieve logs of the given service

        Args:
            details: show service context and extra details provided to logs
            follow: return the logs as a stream.
            stdout: return logs from stdout
            stderr: return logs from stderr
            since: return logs since this time, as a UNIX timestamp
            timestamps: add timestamps to every log line
            is_tty: the service has a pseudo-TTY allocated
            tail: only return this number of log lines
                  from the end of the logs, specify as an integer
                  or `all` to output all log lines.


        """
        if stdout is False and stderr is False:
            raise TypeError("Need one of stdout or stderr")

        params = {
            "details": details,
            "follow": follow,
            "stdout": stdout,
            "stderr": stderr,
            "since": since,
            "timestamps": timestamps,
            "tail": tail
        }

        response = await self.docker._query(
            "services/{service_id}/logs".format(service_id=service_id),
            method="GET",
            params=params
        )
        return (await multiplexed_result(response, follow, is_tty=is_tty))



import asyncio
import datetime
import io
import os
import sys
import tarfile
import time

import aiohttp
import pytest

from aiodocker.docker import Docker


@pytest.mark.asyncio
async def test_autodetect_host(monkeypatch):
    docker = Docker()
    if 'DOCKER_HOST' in os.environ:
        if (os.environ['DOCKER_HOST'].startswith('http://') or
                os.environ['DOCKER_HOST'].startswith('https://') or
                os.environ['DOCKER_HOST'].startswith('tcp://')):
            assert docker.docker_host == os.environ['DOCKER_HOST']
        else:
            assert docker.docker_host == 'unix://localhost'
    else:
        # assuming that docker daemon is installed locally.
        assert docker.docker_host is not None
    await docker.close()


@pytest.mark.asyncio
async def test_connect_invalid_unix_socket():
    docker = Docker('unix:///var/run/does-not-exist-docker.sock')
    assert isinstance(docker.connector, aiohttp.connector.UnixConnector)
    with pytest.raises(aiohttp.ClientOSError):
        await docker.containers.list()
    await docker.close()


@pytest.mark.asyncio
async def test_connect_envvar(monkeypatch):
    monkeypatch.setenv('DOCKER_HOST',
                       'unix:///var/run/does-not-exist-docker.sock')
    docker = Docker()
    assert isinstance(docker.connector, aiohttp.connector.UnixConnector)
    assert docker.docker_host == 'unix://localhost'
    with pytest.raises(aiohttp.ClientOSError):
        await docker.containers.list()
    await docker.close()

    monkeypatch.setenv('DOCKER_HOST', 'http://localhost:9999')
    docker = Docker()
    assert isinstance(docker.connector, aiohttp.TCPConnector)
    assert docker.docker_host == 'http://localhost:9999'
    with pytest.raises(aiohttp.ClientOSError):
        await docker.containers.list()
    await docker.close()


@pytest.mark.asyncio
async def test_connect_with_connector(monkeypatch):
    connector = aiohttp.BaseConnector()
    docker = Docker(connector=connector)
    assert docker.connector == connector
    await docker.close()


@pytest.mark.asyncio
async def test_container_lifecycles(docker, testing_images):
    containers = await docker.containers.list(all=True)
    orig_count = len(containers)

    config = {
        "Cmd": ["/bin/ls"],
        "Image": "alpine:latest",
        "AttachStdin": False,
        "AttachStdout": False,
        "AttachStderr": False,
        "Tty": False,
        "OpenStdin": False,
    }

    my_containers = []
    for i in range(3):
        container = await docker.containers.create(config=config)
        assert container
        my_containers.append(container)

    containers = await docker.containers.list(all=True)
    assert len(containers) == orig_count + 3

    f_container = containers[0]
    await f_container.start()
    await f_container.show()

    for container in my_containers:
        await container.delete(force=True)

    containers = await docker.containers.list(all=True)
    assert len(containers) == orig_count


@pytest.mark.asyncio
@pytest.mark.skipif(sys.platform == 'darwin',
                    reason="Docker for Mac has a bug with websocket")
async def test_stdio_stdin(docker, testing_images, shell_container):
    # echo of the input.
    ws = await shell_container.websocket(stdin=True, stdout=True, stream=True)
    await ws.send_str('echo hello world\n')
    output = b''
    found = False
    try:
        # collect the websocket outputs for at most 2 secs until we see the
        # output.
        with aiohttp.Timeout(2):
            while True:
                output += await ws.receive_bytes()
                if b"echo hello world\r\n" in output:
                    found = True
                    break
    except asyncio.TimeoutError:
        pass
    await ws.close()
    if not found:
        found = b"echo hello world\r\n" in output
    assert found

    # cross-check with container logs.
    log = []
    found = False

    try:
        # collect the logs for at most 2 secs until we see the output.
        stream = await shell_container.log(stdout=True, follow=True)
        with aiohttp.Timeout(2):
            async for s in stream:
                log.append(s)
                if "hello world\r\n" in s:
                    found = True
                    break
    except asyncio.TimeoutError:
        pass
    if not found:
        output = ''.join(log)
        output.strip()
        found = "hello world" in output.split('\r\n')
    assert found


@pytest.mark.asyncio
async def test_wait_timeout(docker, testing_images, shell_container):
    t1 = datetime.datetime.now()
    with pytest.raises(asyncio.TimeoutError):
        await shell_container.wait(timeout=0.5)
    t2 = datetime.datetime.now()
    delta = t2 - t1
    assert delta.total_seconds() < 5


@pytest.mark.asyncio
async def test_put_archive(docker, testing_images):
    config = {
        "Cmd": ["cat", "/tmp/bar/foo.txt"],
        "Image": "alpine:latest",
        "AttachStdin": False,
        "AttachStdout": False,
        "AttachStderr": False,
        "Tty": False,
        "OpenStdin": False
    }

    file_data = b"hello world"
    file_like_object = io.BytesIO()
    tar = tarfile.open(fileobj=file_like_object, mode='w')

    info = tarfile.TarInfo(name='bar')
    info.type = tarfile.DIRTYPE
    info.mode = 0o755
    info.mtime = time.time()
    tar.addfile(tarinfo=info)

    tarinfo = tarfile.TarInfo(name="bar/foo.txt")
    tarinfo.size = len(file_data)
    tar.addfile(tarinfo, io.BytesIO(file_data))
    tar.list()
    tar.close()

    container = await docker.containers.create_or_replace(
        config=config,
        name='aiodocker-testing-archive')
    await container.put_archive(
        path='/tmp',
        data=file_like_object.getvalue())
    await container.start()
    await container.wait(timeout=5)

    output = await container.log(stdout=True, stderr=True)
    assert output[0] == "hello world"

    await container.delete(force=True)


@pytest.mark.asyncio
async def test_get_archive(docker, testing_images):
    config = {
        "Cmd": ["ash", "-c", "echo 'test' > /tmp/foo.txt"],
        "Image": "alpine:latest",
        "AttachStdin": False,
        "AttachStdout": False,
        "AttachStderr": False,
        "Tty": True,
        "OpenStdin": False
    }

    container = await docker.containers.create_or_replace(
        config=config,
        name='aiodocker-testing-get-archive')
    await container.start()
    tar_archive = await container.get_archive('/tmp/foo.txt')

    assert tar_archive is not None
    assert len(tar_archive.members) == 1
    foo_file = tar_archive.extractfile('foo.txt')
    assert foo_file.read() == b'test\n'
    await container.delete(force=True)


@pytest.mark.asyncio
async def test_port(docker, testing_images, redis_container):
    port = await redis_container.port(6379)
    assert port


@pytest.mark.asyncio
async def test_events(docker, testing_images, event_loop):
    subscriber = docker.events.subscribe()

    # Do some stuffs to generate events.
    config = {
        "Cmd": ["/bin/ash"],
        "Image": "alpine:latest",
    }
    container = await docker.containers.create_or_replace(
        config=config,
        name='aiodocker-testing-temp')
    await container.start()
    await container.delete(force=True)

    events_occurred = []
    while True:
        try:
            with aiohttp.Timeout(0.2):
                event = await subscriber.get()
            if event['Actor']['ID'] == container._id:
                events_occurred.append(event['Action'])
        except asyncio.TimeoutError:
            # no more events
            break
        except asyncio.CancelledError:
            break

    assert events_occurred == ['create', 'start', 'kill', 'die', 'destroy']

    await docker.events.stop()



from typing import Any, Dict, List
from .utils import clean_filters


class DockerSwarmNodes(object):
    def __init__(self, docker):
        self.docker = docker

    async def list(self, *, filters: Dict=None) -> List[Dict]:
        """
        Return a list of swarm's nodes.

        Args:
            filters: a dict with a list of filters

        Available filters:
            id=<node id>
            label=<engine label>
            membership=(accepted|pending)`
            name=<node name>
            role=(manager|worker)`
        """

        params = {"filters": clean_filters(filters)}

        response = await self.docker._query_json(
            "nodes",
            method='GET',
            params=params
        )

        return response

    async def inspect(self, *, node_id: str) -> Dict[str, Any]:
        """
        Inspect a node

        Args:
            node_id: The ID or name of the node
        """

        response = await self.docker._query_json(
            "nodes/{node_id}".format(node_id=node_id),
            method="GET",
        )
        return response

    async def update(
        self, *, node_id: str, version: int, spec: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Update the spec of a node.

        Args:
            node_id: The ID or name of the node
            version: version number of the node being updated
            spec: fields to be updated
        """

        params = {"version": version}

        if "Role" in spec:
            assert spec['Role'] in {"worker", "manager"}

        if "Availability" in spec:
            assert spec['Availability'] in {"active", "pause", "drain"}

        response = await self.docker._query_json(
            "nodes/{node_id}/update".format(node_id=node_id),
            method="POST",
            params=params,
            data=spec
        )
        return response

    async def remove(
        self,
        *,
        node_id: str,
        force: bool=False
    ) -> Dict[str, Any]:
        """
        Remove a node from a swarm.

        Args:
            node_id: The ID or name of the node
        """

        params = {"force": force}

        response = await self.docker._query_json(
            "nodes/{node_id}".format(node_id=node_id),
            method="DELETE",
            params=params,
        )
        return response



from collections import ChainMap
import warnings

import aiohttp

from .channel import Channel


class DockerLog:
    def __init__(self, docker, container):
        self.docker = docker
        self.channel = Channel()
        self.container = container
        self.response = None

    def listen(self):
        warnings.warn("use subscribe() method instead",
                      DeprecationWarning, stacklevel=2)
        return self.channel.subscribe()

    def subscribe(self):
        return self.channel.subscribe()

    async def run(self, **params):
        if self.response:
            warnings.warn("already running",
                          RuntimeWarning, stackelevel=2)
            return
        forced_params = {
            'follow': True,
        }
        default_params = {
            'stdout': True,
            'stderr': True,
        }
        params = ChainMap(forced_params, params, default_params)
        try:
            self.response = await self.docker._query(
                'containers/{self.container._id}/logs'.format(self=self),
                params=params,
            )
            while True:
                msg = await self.response.content.readline()
                if not msg:
                    break
                await self.channel.publish(msg)
        except (aiohttp.ClientConnectionError,
                aiohttp.ServerDisconnectedError):
            pass
        finally:
            # signal termination to subscribers
            await self.channel.publish(None)
            try:
                await self.response.release()
            except Exception:
                pass
            self.response = None

    async def stop(self):
        if self.response:
            await self.response.release()



from io import BytesIO

import pytest
from aiodocker import utils
from aiodocker.exceptions import DockerError


@pytest.mark.asyncio
async def test_build_from_remote_file(docker, random_name,
                                      requires_api_version):

    requires_api_version("v1.28",
                         "TODO: test disabled because it fails on "
                         "API version 1.27, this should be fixed")

    remote = ("https://raw.githubusercontent.com/aio-libs/"
              "aiodocker/master/tests/docker/Dockerfile")

    tag = "{}:1.0".format(random_name())
    params = {'tag': tag, 'remote': remote}
    await docker.images.build(**params)

    image = await docker.images.get(tag)
    assert image


@pytest.mark.asyncio
async def test_build_from_remote_tar(docker, random_name):
    remote = ("https://github.com/aio-libs/aiodocker/"
              "raw/master/tests/docker/docker_context.tar")

    tag = "{}:1.0".format(random_name())
    params = {'tag': tag, 'remote': remote}
    await docker.images.build(**params)

    image = await docker.images.get(tag)
    assert image


@pytest.mark.asyncio
async def test_history(docker):
    name = "alpine:latest"
    history = await docker.images.history(name=name)
    assert history


@pytest.mark.asyncio
async def test_list_images(docker):
    name = "alpine:latest"
    images = await docker.images.list(filter=name)
    assert len(images) == 1


@pytest.mark.asyncio
async def test_tag_image(docker, random_name):
    name = "alpine:latest"
    repository = random_name()
    await docker.images.tag(name=name, repo=repository, tag="1.0")
    await docker.images.tag(name=name, repo=repository, tag="2.0")
    image = await docker.images.get(name)
    assert len([x for x in image['RepoTags'] if x.startswith(repository)]) == 2


@pytest.mark.asyncio
async def test_push_image(docker):
    name = "alpine:latest"
    repository = "localhost:5000/image"
    await docker.images.tag(name=name, repo=repository)
    await docker.images.push(name=repository)


@pytest.mark.asyncio
async def test_delete_image(docker):
    name = "alpine:latest"
    repository = "localhost:5000/image"
    await docker.images.tag(name=name, repo=repository)
    assert await docker.images.get(repository)
    await docker.images.delete(name=repository)
    images = await docker.images.list(filter=repository)
    assert len(images) == 0


@pytest.mark.asyncio
async def test_not_existing_image(docker, random_name):
    name = "{}:latest".format(random_name())
    with pytest.raises(DockerError) as excinfo:
        await docker.images.get(name=name)
    assert excinfo.value.status == 404


@pytest.mark.asyncio
async def test_pull_image(docker):
    name = "alpine:latest"
    image = await docker.images.get(name=name)
    assert image


@pytest.mark.asyncio
async def test_build_from_tar(docker, random_name):
    name = "{}:latest".format(random_name())
    dockerfile = '''
    # Shared Volume
    FROM alpine:latest
    VOLUME /data
    CMD ["/bin/sh"]
    '''
    f = BytesIO(dockerfile.encode('utf-8'))
    tar_obj = utils.mktar_from_dockerfile(f)
    await docker.images.build(fileobj=tar_obj, encoding="gzip", tag=name)
    tar_obj.close()
    image = await docker.images.get(name=name)
    assert image


@pytest.mark.asyncio
async def test_pups_image_auth(docker):
    name = "alpine:latest"
    await docker.images.pull(from_image=name)
    repository = "localhost:5001/image:latest"
    image, tag = repository.rsplit(':', 1)
    registry_addr, image_name = image.split('/', 1)
    await docker.images.tag(name=name, repo=image, tag=tag)

    auth_config = {'username': "testuser", 'password': "testpassword"}

    await docker.images.push(name=repository, tag=tag, auth=auth_config)

    await docker.images.delete(name=repository)
    await docker.images.pull(repository,
                             auth={"auth": "dGVzdHVzZXI6dGVzdHBhc3N3b3Jk"})

    await docker.images.get(repository)
    await docker.images.delete(name=repository)

    # Now compose_auth_header automatically parse and rebuild
    # the encoded value if required.
    await docker.pull(repository,
                      auth="dGVzdHVzZXI6dGVzdHBhc3N3b3Jk")
    with pytest.raises(ValueError):
        # The repository arg must include the registry address.
        await docker.pull("image:latest",
                          auth={"auth": "dGVzdHVzZXI6dGVzdHBhc3N3b3Jk"})
    await docker.pull(repository,
                      auth={"auth": "dGVzdHVzZXI6dGVzdHBhc3N3b3Jk"})
    await docker.images.get(repository)



import sys
import types
import asyncio
import struct
import aiohttp

from . import constants
from aiodocker.utils import _DecodeHelper


class MultiplexedResult:
    def __init__(self, response, raw):
        self._response = response

        self._gen = self.fetch
        if raw:
            self._gen = self.fetch_raw

    def __aiter__(self):
        return self

    async def __anext__(self):
        response = await self._gen()

        if not response:
            await self._close()
            raise StopAsyncIteration

        return response

    if sys.version_info <= (3, 5, 2):
        __aiter__ = asyncio.coroutine(__aiter__)

    @types.coroutine
    def fetch(self):
        while True:

            try:
                hdrlen = constants.STREAM_HEADER_SIZE_BYTES
                header = yield from self._response.content.readexactly(hdrlen)

                _, length = struct.unpack('>BxxxL', header)
                if not length:
                    continue

                data = yield from self._response.content.readexactly(length)

            except (aiohttp.ClientConnectionError,
                    aiohttp.ServerDisconnectedError,
                    asyncio.IncompleteReadError):
                break
            return data

    @types.coroutine
    def fetch_raw(self):
        chunk = self._response.content.iter_chunked(1024).__aiter__()
        while True:
            try:
                data = yield from chunk.__anext__()
            except StopAsyncIteration:
                break
            return data

    async def _close(self):
        await self._response.release()


async def multiplexed_result(response, follow=False, is_tty=False,
                             encoding='utf-8'):

    # if is_tty is True you get a raw output
    log_stream = MultiplexedResult(response, raw=is_tty)

    if follow:
        return _DecodeHelper(log_stream, encoding=encoding)
    else:
        d = []
        async for piece in _DecodeHelper(log_stream, encoding=encoding):
            d.append(piece.strip())
        return d



import json


class DockerVolumes:
    def __init__(self, docker):
        self.docker = docker

    async def list(self):
        data = await self.docker._query_json("volumes")
        return data

    async def create(self, config):
        config = json.dumps(config, sort_keys=True).encode('utf-8')
        data = await self.docker._query_json(
            "volumes/create",
            method="POST",
            data=config,
        )
        return DockerVolume(self.docker, data['Name'])


class DockerVolume:
    def __init__(self, docker, name):
        self.docker = docker
        self.name = name

    async def show(self):
        data = await self.docker._query_json(
            "volumes/{self.name}".format(self=self)
        )
        return data

    async def delete(self):
        response = await self.docker._query(
            "volumes/{self.name}".format(self=self),
            method="DELETE",
        )
        await response.release()
        return



import asyncio
from collections import ChainMap
import datetime as dt
import warnings

from .channel import Channel
from .jsonstream import json_stream_result
from .utils import human_bool


class DockerEvents:
    def __init__(self, docker):
        self.docker = docker
        self.channel = Channel()
        self.json_stream = None
        self.task = None

    def listen(self):
        warnings.warn("use subscribe() method instead",
                      DeprecationWarning, stacklevel=2)
        return self.channel.subscribe()

    def subscribe(self, *, create_task=True):
        """Subscribes to the Docker events channel. Use the keyword argument
        create_task=False to prevent automatically spawning the background
        tasks that listen to the events.

        This function returns a ChannelSubscriber object.
        """
        if create_task and not self.task:
            self.task = asyncio.ensure_future(self.run())
        return self.channel.subscribe()

    def _transform_event(self, data):
        if 'time' in data:
            data['time'] = dt.datetime.fromtimestamp(data['time'])
        return data

    async def run(self, **params):
        """
        Query the events endpoint of the Docker daemon.

        Publish messages inside the asyncio queue.
        """
        if self.json_stream:
            warnings.warn("already running",
                          RuntimeWarning, stackelevel=2)
            return
        forced_params = {
            'stream': True,
        }
        params = ChainMap(forced_params, params)
        try:
            # timeout has to be set to 0, None is not passed
            # Otherwise after 5 minutes the client
            # will close the connection
            # http://aiohttp.readthedocs.io/en/stable/client_reference.html#aiohttp.ClientSession.request
            response = await self.docker._query(
                "events",
                method="GET",
                params=params,
                timeout=0
            )
            self.json_stream = await json_stream_result(
                response,
                self._transform_event,
                human_bool(params['stream']),
            )
            try:
                async for data in self.json_stream:
                    await self.channel.publish(data)
            finally:
                if self.json_stream is not None:
                    await self.json_stream._close()
                self.json_stream = None
        finally:
            # signal termination to subscribers
            await self.channel.publish(None)

    async def stop(self):
        if self.json_stream is not None:
            await self.json_stream._close()
        if self.task:
            self.task.cancel()
            try:
                await self.task
            except asyncio.CancelledError:
                pass



class DockerError(Exception):

    def __init__(self, status, data, *args):
        super().__init__(*args)
        self.status = status
        self.message = data['message']

    def __repr__(self):
        return 'DockerError({self.status}, {self.message!r})'.format(self=self)

    def __str__(self):
        return 'DockerError({self.status}, {self.message!r})'.format(self=self)


class DockerContainerError(DockerError):

    def __init__(self, status, data, container_id, *args):
        super().__init__(status, data, *args)
        self.container_id = container_id

    def __repr__(self):
        return ('DockerContainerError('
                '{self.status}, {self.message!r}, '
                '{self.container_id!r})').format(self=self)

    def __str__(self):
        return ('DockerContainerError('
                '{self.status}, {self.message!r}, '
                '{self.container_id!r})').format(self=self)



#!/usr/bin/env python

import re
from pathlib import Path

from setuptools import setup

with (Path(__file__).parent / 'aiodocker' / '__init__.py').open() as fp:
    try:
        version = re.findall(r"^__version__ = '([^']+)'\r?$",
                             fp.read(), re.M)[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')


long_description = open('README.rst').read() + open('CHANGES.rst').read()


with open('./requirements/base.txt') as test_reqs_txt:
    requirements = list(iter(test_reqs_txt))


with open('./requirements/test.txt') as test_reqs_txt:
    test_requirements = list(iter(test_reqs_txt))


setup(
    name="aiodocker",
    version=version,
    author="Paul Tagliamonte",
    author_email="paultag@debian.org",
    long_description=long_description,
    description="Docker API client for asyncio",
    license="Apache 2",
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Framework :: AsyncIO',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
        'Framework :: AsyncIO',
    ],
    url="https://github.com/aio-libs/aiodocker",
    platforms=['any'],
    packages=[
        'aiodocker',
    ],
    python_requires='>=3.5',
    install_requires=requirements,
    tests_require=test_requirements,
    test_suite="tests",
)



import json
import tarfile

from .exceptions import DockerError, DockerContainerError
from .jsonstream import json_stream_result
from .multiplexed import multiplexed_result
from .utils import identical, parse_result

from .logs import DockerLog


class DockerContainers(object):
    def __init__(self, docker):
        self.docker = docker

    async def list(self, **kwargs):
        data = await self.docker._query_json(
            "containers/json",
            method='GET',
            params=kwargs
        )
        return [DockerContainer(self.docker, **x) for x in data]

    async def create_or_replace(self, name, config):
        container = None

        try:
            container = await self.get(name)
            if not identical(config, container._container):
                running = container._container.get(
                    "State", {}).get("Running", False)
                if running:
                    await container.stop()
                await container.delete()
                container = None
        except DockerError:
            pass

        if container is None:
            container = await self.create(config, name=name)

        return container

    async def create(self, config, *, name=None):
        url = "containers/create"

        config = json.dumps(config, sort_keys=True).encode('utf-8')
        kwargs = {}
        if name:
            kwargs['name'] = name
        data = await self.docker._query_json(
            url,
            method='POST',
            data=config,
            params=kwargs
        )
        return DockerContainer(self.docker, id=data['Id'])

    async def run(self, config, *, name=None):
        """
        Create and start a container.

        If container.start() will raise an error the exception will contain
        a `container_id` attribute with the id of the container.
        """
        try:
            container = await self.create(config, name=name)
        except DockerError as err:
            # image not find, try pull it
            if err.status == 404 and 'Image' in config:
                await self.docker.pull(config['Image'])
                container = await self.create(config, name=name)
            else:
                raise err

        try:
            await container.start()
        except DockerError as err:
            raise DockerContainerError(
                err.status,
                {"message": err.message},
                container['id'])

        return container

    async def get(self, container, **kwargs):
        data = await self.docker._query_json(
            "containers/{container}/json".format(container=container),
            method='GET',
            params=kwargs
        )
        return DockerContainer(self.docker, **data)

    def container(self, container_id, **kwargs):
        data = {
            'id': container_id
        }
        data.update(kwargs)
        return DockerContainer(self.docker, **data)


class DockerContainer:
    def __init__(self, docker, **kwargs):
        self.docker = docker
        self._container = kwargs
        self._id = self._container.get("id", self._container.get(
            "ID", self._container.get("Id")))
        self.logs = DockerLog(docker, self)

    async def log(self, *, stdout=False, stderr=False, follow=False, **kwargs):
        if stdout is False and stderr is False:
            raise TypeError("Need one of stdout or stderr")

        params = {
            "stdout": stdout,
            "stderr": stderr,
            "follow": follow,
        }
        params.update(kwargs)

        inspect_info = await self.show()
        is_tty = inspect_info['Config']['Tty']

        response = await self.docker._query(
            "containers/{self._id}/logs".format(self=self),
            method='GET',
            params=params,
        )
        return await multiplexed_result(response, follow, is_tty=is_tty)

    async def copy(self, resource, **kwargs):
        # TODO: this is deprecated, use get_archive instead
        request = json.dumps({
            "Resource": resource,
        }, sort_keys=True).encode('utf-8')
        data = await self.docker._query(
            "containers/{self._id}/copy".format(self=self),
            method='POST',
            data=request,
            headers={"content-type": "application/json"},
            params=kwargs
        )
        return data

    async def get_archive(self, path: str) -> tarfile.TarFile:
        response = await self.docker._query(
            "containers/{self._id}/archive".format(self=self),
            method='GET',
            params={'path': path}
        )
        data = await parse_result(response)
        return data

    async def put_archive(self, path, data):
        response = await self.docker._query(
            "containers/{self._id}/archive".format(self=self),
            method='PUT',
            data=data,
            headers={"content-type": "application/json"},
            params={'path': path}
        )
        data = await parse_result(response)
        return data

    async def show(self, **kwargs):
        data = await self.docker._query_json(
            "containers/{self._id}/json".format(self=self),
            method='GET',
            params=kwargs
        )
        self._container = data
        return data

    async def stop(self, **kwargs):
        response = await self.docker._query(
            "containers/{self._id}/stop".format(self=self),
            method='POST',
            params=kwargs
        )
        await response.release()
        return

    async def start(self, **kwargs):
        response = await self.docker._query(
            "containers/{self._id}/start".format(self=self),
            method='POST',
            headers={"content-type": "application/json"},
            data=kwargs
        )
        await response.release()
        return

    async def kill(self, **kwargs):
        response = await self.docker._query(
            "containers/{self._id}/kill".format(self=self),
            method='POST',
            params=kwargs
        )
        await response.release()
        return

    async def wait(self, *, timeout=None, **kwargs):
        data = await self.docker._query_json(
            "containers/{self._id}/wait".format(self=self),
            method='POST',
            params=kwargs,
            timeout=timeout,
        )
        return data

    async def delete(self, **kwargs):
        response = await self.docker._query(
            "containers/{self._id}".format(self=self),
            method='DELETE',
            params=kwargs
        )
        await response.release()
        return

    async def websocket(self, **params):
        path = "containers/{self._id}/attach/ws".format(self=self)
        ws = await self.docker._websocket(path, **params)
        return ws

    async def port(self, private_port):
        if 'NetworkSettings' not in self._container:
            await self.show()

        private_port = str(private_port)
        h_ports = None

        # Port settings is None when the container is running with
        # network_mode=host.
        port_settings = self._container.get('NetworkSettings', {}).get('Ports')
        if port_settings is None:
            return None

        if '/' in private_port:
            return port_settings.get(private_port)

        h_ports = port_settings.get(private_port + '/tcp')
        if h_ports is None:
            h_ports = port_settings.get(private_port + '/udp')

        return h_ports

    async def stats(self, *, stream=True):
        if stream:
            response = await self.docker._query(
                "containers/{self._id}/stats".format(self=self),
                params={'stream': '1'},
            )
            return (await json_stream_result(response))
        else:
            data = await self.docker._query_json(
                "containers/{self._id}/stats".format(self=self),
                params={'stream': '0'},
            )
            return data

    def __getitem__(self, key):
        return self._container[key]

    def __hasitem__(self, key):
        return key in self._container
